@@ -4,24 +4,25 @@ use strict;
 use warnings;
 use autodie;
 use feature qw(say);
+
+#use FindBin qw($Bin);
+use Cwd                   qw(abs_path);
 use File::Spec::Functions qw(catdir catfile);
+use File::Basename        qw(dirname);
 use Data::Dumper;
 use Path::Tiny;
 use File::Basename;
-use File::ShareDir::ProjectDistDir;
 use List::Util qw(any uniq);
+use Carp       qw(confess);
 use XML::Fast;
 use Moo;
 use Types::Standard qw(Str Int Num Enum ArrayRef Undef);
-use File::ShareDir::ProjectDistDir qw(dist_dir);
 
 #use Devel::Size     qw(size total_size);
-use Convert::Pheno::IO::CSVHandler;
-use Convert::Pheno::IO::FileIO;
+use Convert::Pheno::CSV;
+use Convert::Pheno::IO;
 use Convert::Pheno::SQLite;
 use Convert::Pheno::Mapping;
-use Convert::Pheno::CSV;
-use Convert::Pheno::RDF qw(do_bff2jsonld do_pxf2jsonld);
 use Convert::Pheno::OMOP;
 use Convert::Pheno::PXF;
 use Convert::Pheno::BFF;
@@ -29,24 +30,15 @@ use Convert::Pheno::CDISC;
 use Convert::Pheno::REDCap;
 
 use Exporter 'import';
-our @EXPORT =
-  qw($VERSION io_yaml_or_json omop2bff_stream_processing share_dir);    # Symbols imported by default
+our @EXPORT = qw($VERSION io_yaml_or_json omop2bff_stream_processing);    # Symbols imported by default
 
 #our @EXPORT_OK = qw(foo bar);       # Symbols imported by request
 
 use constant DEVEL_MODE => 0;
 
-# Personalize warn and die functions
-$SIG{__WARN__} = sub { warn "Warn: ", @_ };
-$SIG{__DIE__}  = sub { die "Error: ", @_ };
-
 # Global variables:
-our $VERSION   = '0.26_1';
-our $share_dir = dist_dir('Convert-Pheno');
-
-# SQLite database
-my @all_sqlites       = qw(ncit icd10 ohdsi cdisc omim hpo);
-my @non_ohdsi_sqlites = qw(ncit icd10 cdisc omim hpo);
+our $VERSION  = '0.01';
+our $lib_path = dirname( abs_path(__FILE__) );
 
 ############################################
 # Start declaring attributes for the class #
@@ -82,26 +74,10 @@ has min_text_similarity_score => (
 
 has username => (
 
-    #default => ( $ENV{LOGNAME} || $ENV{USER} || getpwuid($<) ) , # getpwuid not implemented in Windows
-    default => $ENV{'LOGNAME'}
-      || $ENV{'USER'}
-      || $ENV{'USERNAME'}
-      || 'dummy-user',
-    is     => 'ro',
-    coerce => sub {
-        $_[0] // ( $ENV{'LOGNAME'}
-              || $ENV{'USER'}
-              || $ENV{'USERNAME'}
-              || 'dummy-user' );
-    },
-    isa => Str
-);
-
-has id => (
-    default => time . substr( "00000$$", -5 ),
+    default => ( $ENV{LOGNAME} || $ENV{USER} || getpwuid($<) ),
     is      => 'ro',
     coerce  => sub {
-        $_[0] // time . substr( "00000$$", -5 );
+        $_[0] // ( $ENV{LOGNAME} || $ENV{USER} || getpwuid($<) );
     },
     isa => Str
 );
@@ -113,18 +89,15 @@ has max_lines_sql => (
     isa     => Int
 );
 
-has 'omop_tables' => (
+has omop_tables => (
+
+    # Table <CONCEPT> is always required
     default => sub { [@omop_essential_tables] },
     coerce  => sub {
-        my $tables = shift;
-
-        # If tables are provided, process them; otherwise, use default essential tables
-        $tables =
-          @$tables
-          ? [ uniq( map { uc($_) } ( 'CONCEPT', 'PERSON', @$tables ) ) ]
+        @{ $_[0] }
+          ? $_[0] =
+          [ map { uc($_) } ( uniq( @{ $_[0] }, 'CONCEPT', 'PERSON' ) ) ]
           : \@omop_essential_tables;
-
-        return $tables;
     },
     is  => 'rw',
     isa => ArrayRef
@@ -133,10 +106,10 @@ has 'omop_tables' => (
 has exposures_file => (
 
     default =>
-      catfile( $share_dir, 'db', 'concepts_candidates_2_exposure.csv' ),
+      catfile( $lib_path, '../../db/concepts_candidates_2_exposure.csv' ),
     coerce => sub {
         $_[0]
-          // catfile( $share_dir, 'db', 'concepts_candidates_2_exposure.csv' );
+          // catfile( $lib_path, '../../db/concepts_candidates_2_exposure.csv' );
     },
     is  => 'ro',
     isa => Str
@@ -160,17 +133,20 @@ has [qw /data method/] => ( is => 'rw' );
 # End declaring attributes for the class #
 ##########################################
 
-sub BUILD {
-
-    # BUILD: is an instance method that is called after the object has been constructed but before it is returned to the caller.
-    # BUILDARGS is a class method that is responsible for processing the arguments passed to the constructor (new) and returning a hash reference of attributes that will be used to initialize the object.
-    my $self = shift;
-    $self->{databases} =
-      $self->{ohdsi_db} ? \@all_sqlites : \@non_ohdsi_sqlites;
-}
-
 # NB: In general, we'll only display terms that exist and have content
 
+#############
+#############
+#  PXF2BFF  #
+#############
+#############
+
+sub pxf2bff {
+
+    # <array_dispatcher> will deal with JSON arrays
+    return array_dispatcher(shift);
+}
+
 #############
 #############
 #  BFF2PXF  #
@@ -179,47 +155,8 @@ sub BUILD {
 
 sub bff2pxf {
 
-    my $self = shift;
     # <array_dispatcher> will deal with JSON arrays
-    return $self->array_dispatcher;
-}
-
-#############
-#############
-#  BFF2CSV  #
-#############
-#############
-
-sub bff2csv {
-
-    my $self = shift;
-    # <array_dispatcher> will deal with JSON arrays
-    return $self->array_dispatcher;
-}
-
-#############
-#############
-# BFF2JSONF #
-#############
-#############
-
-sub bff2jsonf {
-
-    my $self = shift;
-    # <array_dispatcher> will deal with JSON arrays
-    return $self->array_dispatcher;
-}
-##############
-##############
-# BFF2JSONLD #
-##############
-##############
-
-sub bff2jsonld {
-
-    my $self = shift;
-    # <array_dispatcher> will deal with JSON arrays
-    return $self->array_dispatcher;
+    return array_dispatcher(shift);
 }
 
 ################
@@ -233,29 +170,24 @@ sub redcap2bff {
     my $self = shift;
 
     # Read and load data from REDCap export
-    my $data = read_csv( { in => $self->{in_file}, sep => $self->{sep} } );
-    my $data_redcap_dict = read_redcap_dict_file(
-        {
-            redcap_dictionary => $self->{redcap_dictionary},
-        }
-    );
-    my $data_mapping_file = read_mapping_file(
+    my $data = read_csv( { in => $self->{in_file}, sep => undef } );
+    my ( $data_redcap_dict, $data_mapping_file ) =
+      read_redcap_dict_and_mapping_file(
         {
+            redcap_dictionary    => $self->{redcap_dictionary},
             mapping_file         => $self->{mapping_file},
             self_validate_schema => $self->{self_validate_schema},
             schema_file          => $self->{schema_file}
         }
-    );
+      );
 
     # Load data in $self
-    $self->{data}              = $data;                  # Dynamically adding attributes (setter)
-    $self->{data_redcap_dict}  = $data_redcap_dict;      # Dynamically adding attributes (setter)
-    $self->{data_mapping_file} = $data_mapping_file;     # Dynamically adding attributes (setter)
-    $self->{metaData}          = get_metaData($self);    # Dynamically adding attributes (setter)
-    $self->{convertPheno}      = get_info($self);        # Dynamically adding attributes (setter)
+    $self->{data}              = $data;                 # Dynamically adding attributes (setter)
+    $self->{data_redcap_dict}   = $data_redcap_dict;      # Dynamically adding attributes (setter)
+    $self->{data_mapping_file} = $data_mapping_file;    # Dynamically adding attributes (setter)
 
     # array_dispatcher will deal with JSON arrays
-    return $self->array_dispatcher;
+    return array_dispatcher($self);
 }
 
 ################
@@ -278,7 +210,7 @@ sub redcap2pxf {
     $self->{in_textfile} = 0;            # setter
 
     # Run second iteration
-    return $self->array_dispatcher;
+    return array_dispatcher($self);
 }
 
 ##############
@@ -295,47 +227,58 @@ sub omop2bff {
     # IMPORTANT #
     #############
 
-    # File Size Considerations for Data Processing
+    # SMALL TO MEDIUM FILES < 1M rows
     #
-    # For SMALL TO MEDIUM FILES (< 1M rows):
-    # Commonly, database downsizing for data sharing results in PostgreSQL dumps or CSVs being less than 1 million rows.
-    # With adequate memory (4-16GB), we can efficiently load this data into RAM and effectively consolidate individual data points (e.g., MEASURES, DRUGS).
+    # In many cases, because people are downsizing their DBs for data sharing,
+    # PostgreSQL dumps or CSVs will be < 1M rows.
+    # Providing we have enough memory (4-16GB), we'll able to load data in RAM,
+    # and consolidate individual values (MEASURES, DRUGS, etc.)
+
+    # HUMONGOUS FILES > 1M rows
+    # NB: Interesting read on the topic
+    #     https://www.perlmonks.org/?node_id=1033692
+    # Since we're relying heavily on hashes we need to resort to another strategy(es) to load the data
     #
-    # For HUMONGOUS FILES (> 1M rows):
-    # As we heavily use hashes, larger files necessitate alternative data loading strategies:
+    # * Option A *: Parellel processing - No change in our code
+    #    Without changing the code, we ask the user to create mini-instances (or split CSV's in chunks) and use
+    #    some sort of parallel processing (e.g., GNU parallel, snakemake, HPC, etc.)
+    #    CONS: Concurrent jobs may fail due to SQLite been opened by multiple threads
     #
-    # * Option A: Parallel Processing (No code modification required)
-    #   Users can split their data into smaller chunks or mini-instances, employing parallel processing tools (like GNU parallel, snakemake, HPC, etc.).
-    #   Caveat: SQLite’s limitations with concurrent access by multiple threads.
-    #
-    # * Option B: Data Consolidation at Individual Object Level
+    # * Option B *: Keeping data consolidated at the individual-object level (as we do with small to medium files)
     #   --no-stream
-    #   Two approaches for this:
-    #     a) Externalize the complete hash using DBM:Deep (although it's significantly slower).
-    #     b) Initially dump data as CSV (either by the user or automatically), then sort it (using *nix or SQLite) by 'person_id'.
-    #        This method doesn't substantially help with data consolidation since we still process one table at a time.
+    #   To do this, we have two options:
+    #     a) Externalize (save to file) THE WHOLE HASH w/ DBM:Deep (but it's very slow)
+    #     b) First dump CSV (me or users) and then use *nix to sort by person_id (or loadSQLite and sort there).
+    #   Then, since rows for each individual are adjacent, we can load individual data together. Still,
+    #   we'll by reading one table (e.g. MEASUREMENTS) at a time, thus, this is not relly helping much to consolidate...
     #
-    # * Option C: Line-by-Line File Parsing (One row of CSV/SQL per JSON object) <===== CURRENT IMPLEMENTATION
+    # * Option C *: Parsing files line by line (one row of CSV/SQL per JSON object) <=========== IMPLEMENTED ==========
     #   --stream
-    #   Note: BFF / PXF JSON files serve as intermediate stages. They group data by individual for easier inspection but are ultimately stored in Mongo DB.
-    #   Similar to the genomicVariations issue in B2RI, multiple JSON objects (like MEASUREMENTS, DRUGS) can correspond to a single individual.
-    #   The link is the term "id"
+    #   BFF / PXF JSONs are just intermediate files. It's nice that they contain data grouped by individual
+    #   (for visually inspection and display), but at the end of the day they'll end up in Mongo DB.
+    #   If all entries contain the primary key 'person_id' then it's up to the Beacon v2 API to deal with them.
+    #   It's a similar issue to the one we had with genomicVariations in the B2RI, where a given variant belong to many individuals.
+    #   Here, multiple JSON documents/objects (MEASUREMENTS, DRUGS, etc.) will belong to the same individual.
+    #   Now, since we allow for CSV and SQL as an input, we need to minimize the numer of steps to a minimum.
     #
-    #   Potential Issues and Solutions:
-    #     1. Mandatory <CONCEPT> Table:
-    #        It can be extremely large, potentially consuming all available RAM (e.g., a 735 MB <CONCEPT.csv> with over 5.8 million lines).
-    #        Solutions:
-    #          a) Avoid loading the <CONCEPT> table entirely, using --ohdsi-db instead.
-    #          b) Use a temporary SQLite instance for the <CONCEPT> table.
-    #     2. Reading SQL Dumps Line-by-Line:
-    #        For large SQL dumps (e.g., 20GB), should we convert them into CSV (also ~20GB)?
-    #        Solutions:
-    #          a) Yes, first export required tables to CSV and then proceed.
-    #          b) No, read the PostgreSQL dump twice - first to load specified tables, then the rest.
-    #     3. Streaming Mode Restrictions:
-    #        In --stream mode, --sql2csv is not allowed to prevent excessive space usage and complexity.
+    #   - Problems that may arise:
+    #     1 - <CONCEPT> table is mandatory, but it can be so huge that it takes all RAM memory.
+    #         For instance, <CONCEPT.csv> with 5_808_095 lines = 735 MB
+    #                       <CONCEPT_light.csv> with 5_808_094 lines but only 4 columns = 501 MB
+    #                       Anything more than 2M lines kills a 8GB Ram machine.
+    #         Solutions:
+    #           a) Not loading the table at all and resort to --ohdsi-db
+    #           b) Creating a temporary SQLite instance for <CONCEPT>
+    #     2 - How to read line-by-line from an SQL dump
+    #          If the PostgreSQL dump weights, say, 20GB, do we create CSV tables from it (another ~20GB)?
+    #         Solutions:
+    #           a) Yep, we read @stream_ram_memory_tables and  export the needed tables to CSV and go from there.
+    #           b) Nope, we read PostgreSQL file twice, one time to load @stream_ram_memory_tables
+    #              and the second time to load the remaining TABLES. <=========== IMPLEMENTED ==========
+    #     3 - In --stream mode, do we still allow for --sql2csv? NOPE !!!! <=========== IMPLEMENTED ==========
+    #           We would need to go from functional mode (csv) to filehandles and it will take tons of space.
+    #           Then, --stream and -sql2csv are mutually exclusive.
     #
-    # Further reading on handling large files: https://www.perlmonks.org/?node_id=1033692
 
     # Load variables
     my $data;
@@ -369,20 +312,15 @@ sub omop2bff {
 
         for my $file ( @{ $self->{in_files} } ) {
             my ( $table_name, undef, $ext ) = fileparse( $file, @exts );
-
-            #####################
-            # PostgreSQL export #
-            #####################
-
             if ( $ext =~ m/\.sql/i ) {
 
-                print "> Param: --max-lines-sql = $self->{max_lines_sql}\n" if $self->{verbose};
+                #######################
+                # Loading OMOP tables #
+                #######################
 
                 # --no-stream
                 if ( !$self->{stream} ) {
 
-                    print "> Mode : --no-stream\n\n" if $self->{verbose};
-
                     # We read all tables in memory
                     $data = read_sqldump( { in => $file, self => $self } );
 
@@ -393,28 +331,18 @@ sub omop2bff {
                 # --stream
                 else {
 
-                    print "> Mode : --stream\n\n" if $self->{verbose};
-
                     # We'll ONLY load @stream_ram_memory_tables
                     # in RAM and the other tables as $fh
                     $self->{omop_tables} = [@stream_ram_memory_tables];    # setter
                     $data = read_sqldump( { in => $file, self => $self } );
                 }
 
-                # Misc print
-                print "> Parameter --max-lines-sql set to: $self->{max_lines_sql}\n\n" if $self->{verbose};
-
                 # We keep the filepath for later
                 $filepath = $file;
 
                 # Exit loop
                 last;
             }
-
-            #############
-            # CSV files #
-            #############
-
             else {
 
                 # We'll load all OMOP tables that the user is providing as -iomop
@@ -425,24 +353,21 @@ sub omop2bff {
                   #unless (any { $_ eq $table_name } @{ $omop_main_table->{$omop_version} };
                   unless any { $_ eq $table_name } @omop_essential_tables;    # global
 
-                my $msg =
-                  "Reading <$table_name> and storing it in RAM memory...";
-
                 # --no-stream
                 if ( !$self->{stream} ) {
 
                     # We read all tables in memory
-                    say $msg if ( $self->{verbose} || $self->{debug} );
                     $data->{$table_name} =
-                      read_csv( { in => $file, sep => $self->{sep}, self => $self } );
+                      read_csv( { in => $file, sep => $self->{sep} } );
                 }
 
                 # --stream
                 else {
+                    # We'll ONLY load @stream_ram_memory_tables
+                    # in RAM and the other tables as $fh
                     if ( any { $_ eq $table_name } @stream_ram_memory_tables ) {
-                        say $msg if ( $self->{verbose} || $self->{debug} );
                         $data->{$table_name} =
-                          read_csv( { in => $file, sep => $self->{sep}, self => $self } );
+                          read_csv( { in => $file, sep => $self->{sep} } );
                     }
                     else {
                         push @filepaths, $file;
@@ -455,48 +380,32 @@ sub omop2bff {
     #print Dumper_concise($data) and die;
     #print Dumper_concise($self) and die;
 
-    # *** IMPORTANT ***
-    # ABOUT TABLE <CONCEPT> BEING MANDATORY
-    # Options:
-    # a) MANDATORY <== IMPLEMENTED
-    #    Drawback: High RAM usage with <PERSON> for large tables
-    # b) OPTIONAL (with --ohdsi-db)
-    #    Using external SQLite database is possible, but risks missing ad hoc concept_ids
-    # Note: CSV users might not provide <CONCEPT.csv> with CSVs
-    die "The table <CONCEPT> is missing from the input files\n"
+    # Primarily with CSVs, it can happen that user does not provide <CONCEPT.csv>
+    confess 'We could not find table <CONCEPT> from your input files'
       unless exists $data->{CONCEPT};
 
     # We create a dictionary for $data->{CONCEPT}
-    $self->{data_ohdsi_dict} = convert_table_aoh_to_hoh( $data, 'CONCEPT', $self );   # Dynamically adding attributes (setter)
-
-    # Transform Array of Hashes (AoH) to Hash of Hashes (HoH) for faster computation
-    if ( $self->{stream} ) {
-        $self->{person} = convert_table_aoh_to_hoh( $data, 'PERSON', $self );         # Dynamically adding attributes (setter)
-    }
-
-    # We convert $self->{data}{VISIT_OCCURRENCE} if present
-    if ( exists $data->{VISIT_OCCURRENCE} ) {
-        $self->{visit_occurrence} =
-          convert_table_aoh_to_hoh( $data, 'VISIT_OCCURRENCE', $self );               # Dynamically adding attributes (setter)
-        delete $data->{VISIT_OCCURRENCE};                                      # Anyway, $data->{VISIT_OCCURRENCE} = [] from convert_table_aoh_to_hoh
-    }
+    $self->{data_ohdsi_dic} = transpose_ohdsi_dictionary( $data->{CONCEPT} );  # Dynamically adding attributes (setter)
 
     # We load the allowed concept_id for exposures as hashref (for --no--stream and --stream)
     $self->{exposures} = load_exposures( $self->{exposures_file} );            # Dynamically adding attributes (setter)
 
-    # Now we need to perform a transformation of the data where 'person_id' is one row of data
+    # We transpose $self->{data}{VISIT_OCCURRENCE} if present
+    if ( exists $data->{VISIT_OCCURRENCE} ) {
+        $self->{visit_occurrence} =
+          transpose_visit_occurrence( $data->{VISIT_OCCURRENCE} );             # Dynamically adding attributes (setter)
+        delete $data->{VISIT_OCCURRENCE};
+    }
+
+    # Now we need to perform a tranformation of the data where 'person_id' is one row of data
     # NB: Transformation is due ONLY IN $omop_main_table FIELDS, the rest of the tables are not used
     # The transformation is performed in --no-stream mode
     $self->{data} =
-      $self->{stream} ? $data : transpose_omop_data_structure($self, $data);    # Dynamically adding attributes (setter)
+      $self->{stream} ? $data : transpose_omop_data_structure($data);    # Dynamically adding attributes (setter)
 
     # Giving some memory back to the system
     $data = undef;
 
-    # Adding miscellanea metadata
-    $self->{metaData}     = get_metaData($self);                         # setter
-    $self->{convertPheno} = get_info($self);                             # setter
-
     # --stream
     if ( $self->{stream} ) {
         omop_stream_dispatcher(
@@ -507,7 +416,7 @@ sub omop2bff {
     # --no-stream
     else {
         # array_dispatcher will deal with JSON arrays
-        return $self->array_dispatcher;
+        return array_dispatcher($self);
     }
 }
 
@@ -541,7 +450,7 @@ sub omop2pxf {
         $self->{in_textfile} = 0;            # setter
 
         # Run second iteration
-        return $self->array_dispatcher;
+        return array_dispatcher($self);
 
         # CLI
     }
@@ -570,26 +479,23 @@ sub cdisc2bff {
     my $hash = xml2hash $str, attr => '-', text => '~';
     my $data = cdisc2redcap($hash);
 
-    my $data_redcap_dict = read_redcap_dict_file(
-        {
-            redcap_dictionary => $self->{redcap_dictionary},
-        }
-    );
-    my $data_mapping_file = read_mapping_file(
+    my ( $data_redcap_dict, $data_mapping_file ) =
+      read_redcap_dict_and_mapping_file(
         {
+            redcap_dictionary    => $self->{redcap_dictionary},
             mapping_file         => $self->{mapping_file},
             self_validate_schema => $self->{self_validate_schema},
             schema_file          => $self->{schema_file}
         }
-    );
+      );
 
     # Load data in $self
     $self->{data}              = $data;                 # Dynamically adding attributes (setter)
-    $self->{data_redcap_dict}  = $data_redcap_dict;     # Dynamically adding attributes (setter)
+    $self->{data_redcap_dict}   = $data_redcap_dict;      # Dynamically adding attributes (setter)
     $self->{data_mapping_file} = $data_mapping_file;    # Dynamically adding attributes (setter)
 
     # array_dispatcher will deal with JSON arrays
-    return $self->array_dispatcher;
+    return array_dispatcher($self);
 }
 
 ###############
@@ -612,115 +518,7 @@ sub cdisc2pxf {
     $self->{in_textfile} = 0;            # setter
 
     # Run second iteration
-    return $self->array_dispatcher;
-}
-
-#############
-#############
-#  PXF2BFF  #
-#############
-#############
-
-sub pxf2bff {
-
-    my $self = shift;
-    # <array_dispatcher> will deal with JSON arrays
-    return $self->array_dispatcher;
-}
-
-#############
-#############
-#  CSV2BFF  #
-#############
-#############
-
-sub csv2bff {
-
-    my $self = shift;
-
-    # Read and load data from CSV
-    my $data = read_csv( { in => $self->{in_file}, sep => $self->{sep} } );
-
-    # Read and load mapping file
-    my $data_mapping_file = read_mapping_file(
-        {
-            mapping_file         => $self->{mapping_file},
-            self_validate_schema => $self->{self_validate_schema},
-            schema_file          => $self->{schema_file}
-        }
-    );
-
-    # Load data in $self
-    $self->{data}              = $data;                  # Dynamically adding attributes (setter)
-    $self->{data_mapping_file} = $data_mapping_file;     # Dynamically adding attributes (setter)
-    $self->{metaData}          = get_metaData($self);    # Dynamically adding attributes (setter)
-    $self->{convertPheno}      = get_info($self);        # Dynamically adding attributes (setter)
-
-    # array_dispatcher will deal with JSON arrays
-    return $self->array_dispatcher;
-}
-
-#############
-#############
-#  CSV2PXF  #
-#############
-#############
-
-sub csv2pxf {
-
-    my $self = shift;
-
-    # First iteration: csv2bff
-    $self->{method} = 'csv2bff';    # setter - we have to change the value of attr {method}
-    my $bff = csv2bff($self);       # array
-
-    # Preparing for second iteration: bff2pxf
-    $self->{method}      = 'bff2pxf';    # setter
-    $self->{data}        = $bff;         # setter
-    $self->{in_textfile} = 0;            # setter
-
-    # Run second iteration
-    return $self->array_dispatcher;
-
-}
-
-#############
-#############
-#  PXF2CSV  #
-#############
-#############
-
-sub pxf2csv {
-
-    my $self = shift;
-    # <array_dispatcher> will deal with JSON arrays
-    return $self->array_dispatcher;
-}
-
-#############
-#############
-# PXFJSONF #
-#############
-#############
-
-sub pxf2jsonf {
-
-    my $self = shift;
-    # <array_dispatcher> will deal with JSON arrays
-    return $self->array_dispatcher;
-}
-
-##############
-##############
-# PXF2JSONLD #
-##############
-##############
-
-sub pxf2jsonld {
-
-    my $self = shift;
-    # <array_dispatcher> will deal with JSON arrays
-    return $self->array_dispatcher;
+    return array_dispatcher($self);
 }
 
 ######################
@@ -735,26 +533,17 @@ sub array_dispatcher {
 
     # Load the input data as Perl data structure
     my $in_data =
-      (      $self->{in_textfile}
-          && $self->{method} !~ m/^(redcap2|omop2|cdisc2|csv)/ )
+      ( $self->{in_textfile} && $self->{method} !~ m/^redcap2|^omop2|^cdisc2/ )
       ? io_yaml_or_json( { filepath => $self->{in_file}, mode => 'read' } )
       : $self->{data};
 
     # Define the methods to call (naming 'func' to avoid confussion with $self->{method})
     my %func = (
+        pxf2bff    => \&do_pxf2bff,
         redcap2bff => \&do_redcap2bff,
         cdisc2bff  => \&do_cdisc2bff,
         omop2bff   => \&do_omop2bff,
-        csv2bff    => \&do_csv2bff,
-        csv2pxf    => \&do_csv2pxf,
-        bff2pxf    => \&do_bff2pxf,
-        bff2csv    => \&do_bff2csv,
-        bff2jsonf  => \&do_bff2csv,      # Not a typo, is the same as above
-        bff2jsonld => \&do_bff2jsonld,
-        pxf2bff    => \&do_pxf2bff,
-        pxf2csv    => \&do_pxf2csv,
-        pxf2jsonf  => \&do_pxf2csv,      # Not a typo, is the same as above
-        pxf2jsonld => \&do_pxf2jsonld
+        bff2pxf    => \&do_bff2pxf
     );
 
     # Open connection to SQLlite databases ONCE
@@ -767,13 +556,9 @@ sub array_dispatcher {
         say $fh_out "[";
     }
 
-    # *** IMPORTANT ***
-    # $out_data = Caution with RAM 
-    # We store all in memory and serialize externally
-    # except for omop2bff (larger) that we print to file here (item-by-item)
-    my $out_data;
-
     # Proceed depending if we have an ARRAY or not
+    # NB: Caution with RAM (we store all in memory except for omop2bff)
+    my $out_data;
     if ( ref $in_data eq ref [] ) {
 
         # Print if we have ARRAY
@@ -788,17 +573,15 @@ sub array_dispatcher {
         # In $self->{data} we have all participants data, but,
         # WE DELIBERATELY SEPARATE ARRAY ELEMENTS FROM $self->{data}
 
-        # We void items in $in_data to avoid data duplication in RAM
-        while ( my $item = shift @{ $in_data} ) {         # We want to keep order (!pop)
-
+        for ( @{$in_data} ) {
             $count++;
 
-            # Print info
+            # Print imfo
             say "[$count] ARRAY ELEMENT from $elements" if $self->{debug};
 
             # NB: If we get "null" participants the validator will complain
             # about not having "id" or any other required property
-            my $method_result = $func{ $self->{method} }->( $self, $item );    # Method
+            my $method_result = $func{ $self->{method} }->( $self, $_ );    # Method
 
             # Only proceeding if we got value from method
             if ($method_result) {
@@ -809,12 +592,16 @@ sub array_dispatcher {
                 if ( exists $self->{omop_cli} && $self->{omop_cli} ) {
                     my $out = omop_dispatcher( $self, $method_result );
                     print $fh_out $$out;
-                    print $fh_out ",\n" unless ( $total == $elements || $total == $self->{max_lines_sql} );
+                    print $fh_out ",\n"
+                      unless ( $total == $elements
+                        || $total == $self->{max_lines_sql} );
                 }
 
                 # For the other we have array_ref $out_data and serialize at once
                 else {
                     push @{$out_data}, $method_result;
+
+                    #say total_size($out_data);
                 }
             }
         }
@@ -852,16 +639,13 @@ sub omop_dispatcher {
 
     # omop2bff encode directly
     if ( $self->{method_ori} ne 'omop2pxf' ) {
-
-        # Watch out!! Don't double encode by using utf8 here. $fh is already utf-8!!!
-        #out = JSON::XS->new->utf8->canonical->pretty->encode($method_result);
-        $out = JSON::XS->new->canonical->pretty->encode($method_result);
+        $out = JSON::XS->new->utf8->canonical->pretty->encode($method_result);
     }
 
     # omop2pxf convert to PXF
     else {
         my $pxf = do_bff2pxf( $self, $method_result );
-        $out = JSON::XS->new->canonical->pretty->encode($pxf);
+        $out = JSON::XS->new->utf8->canonical->pretty->encode($pxf);
     }
     chomp $out;
     return \$out;
@@ -875,48 +659,47 @@ sub omop_stream_dispatcher {
     my $filepaths   = $arg->{filepaths};
     my $omop_tables = $self->{prev_omop_tables};
 
-    # Open a SQLite database connection if required
+    # Open connection to SQLite databases ONCE
     open_connections_SQLite($self) if $self->{method} ne 'bff2pxf';
 
-    # Process files based on the input type (CSV or PostgreSQL dump)
-    return @$filepaths
-      ? process_csv_files_stream( $self, $filepaths )
-      : process_sqldump_stream( $self, $filepath, $omop_tables, $self );
-}
+    # First we do transformations from AoH to HoH to speed up the calculation
+    my $person = { map { $_->{person_id} => $_ } @{ $self->{data}{PERSON} } };
 
-sub process_csv_files_stream {
+    # Give back memory to RAM
+    delete $self->{data}{PERSON};
 
-    my ( $self, $filepaths ) = @_;
-    my $person = $self->{person};
-    for my $file (@$filepaths) {
-        say "Processing file ... <$file>" if $self->{verbose};
-        read_csv_stream(
-            {
-                in     => $file,
-                sep    => $self->{sep},
-                self   => $self,
-                person => $person
-            }
-        );
+    # CSVs
+    if (@$filepaths) {
+        for (@$filepaths) {
+            say "Processing file ... <$_>" if $self->{verbose};
+            read_csv_stream(
+                {
+                    in     => $_,
+                    sep    => $self->{sep},
+                    self   => $self,
+                    person => $person
+                }
+            );
+        }
     }
-    return 1;
-}
 
-sub process_sqldump_stream {
+    # PosgreSQL dump
+    else {
 
-    my ( $self, $filepath, $omop_tables ) = @_;
-    my $person = $self->{person};
+        # Now iterate
+        for my $table ( @{$omop_tables} ) {
 
-    # *** IMPORTANT ***
-    # We proceed as we do with CSV, opening the file for every table
-    # With PosgtreSQL.dumps gzipped the overhead means 2x time.
-    for my $table (@$omop_tables) {
-        next if any { $_ eq $table } @stream_ram_memory_tables;
-        say "Processing table <$table> line-by-line..." if $self->{verbose};
-        $self->{omop_tables} = [$table];
-        read_sqldump_stream(
-            { in => $filepath, self => $self, person => $person } );
+            # We already loaded @stream_ram_memory_tables;
+            next if any { $_ eq $table } @stream_ram_memory_tables;
+            say "Processing table ... <$table>" if $self->{verbose};
+            $self->{omop_tables} = [$table];
+            read_sqldump_stream(
+                { in => $filepath, self => $self, person => $person } );
+        }
     }
+
+    # Close connections ONCE
+    close_connections_SQLite($self) unless $self->{method} eq 'bff2pxf';
     return 1;
 }
 
@@ -951,57 +734,35 @@ Convert::Pheno - A module to interconvert common data models for phenotypic data
 
  use Convert::Pheno;
 
- my $my_pxf_json_data = {
-     "phenopacket" => {
-         "id"      => "P0007500",
-         "subject" => {
-             "id"          => "P0007500",
-             "dateOfBirth" => "unknown-01-01T00:00:00Z",
-             "sex"         => "FEMALE"
-         }
-     }
- };
-
- # Create object
- my $convert = Convert::Pheno->new(
-     {
-         data   => $my_pxf_json_data,
-         method => 'pxf2bff'
-     }
- );
-
- # Apply a method
- my $data = $convert->pxf2bff;
+ # Create a new object
+ 
+ my $convert = Convert::Pheno->new($input);
+ 
+ # Apply a method 
+ 
+ my $data = $convert->redcap2bff;
 
 =head1 DESCRIPTION
 
-For a better description, please read the following documentation:
-
-=over
-
-=item General:
-
-L<https://cnag-biomedical-informatics.github.io/convert-pheno>
-
-=item Command-Line Interface:
-
-L<https://github.com/CNAG-Biomedical-Informatics/convert-pheno#readme>
-
-=back
-
 =head1 CITATION
 
 The author requests that any published work that utilizes C<Convert-Pheno> includes a cite to the the following reference:
 
-Rueda, M et al., (2024). Convert-Pheno: A software toolkit for the interconversion of standard data models for phenotypic data. Journal of Biomedical Informatics. L<DOI|https://doi.org/10.1016/j.jbi.2023.104558>
+Rueda, M. "Convert-Pheno: A software toolkit for the interconversion of standard data models for phenotypic data". I<iManuscript in preparation>.
 
 =head1 AUTHOR
 
-Written by Manuel Rueda, PhD. Info about CNAG can be found at L<https://www.cnag.eu>.
+Written by Manuel Rueda, PhD. Info about CNAG can be found at L<https://www.cnag.crg.eu>.
 
 =head1 METHODS
 
-See L<https://cnag-biomedical-informatics.github.io/convert-pheno/use-as-a-module>.
+=head2 COMMON ERRORS AND SOLUTIONS
+
+ * Error message: Foo
+   Solution: Bar
+
+ * Error message: Foo
+   Solution: Bar
 
 =head1 COPYRIGHT
 
